#include <WiFi.h>
#include <HTTPClient.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>
#include <ArduinoJson.h>
#include <string>
#include <ESP32Servo.h>
Servo servo01;
Servo servo02;

#define SCREEN_WIDTH 128 
#define SCREEN_HEIGHT 64

#define OLED_RESET     -1 
Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire, OLED_RESET);

#define MAX_DEVICES 20

const char* ssid = "Qwerty";
const char* password = "12345678";

int servo_pin[2] = { 12, 14 };
int led_pin[2] = { 18, 19 };

int gas_pin = 33;
int temp_pin = 35;
int motor_pin = 32;
const int pingPin = 15;
int inPin = 13;

int gascounter = 0;
int nearsonicounter = 0;
int nearsonicpass = 0;
int gas_array[10]={0,0,0,0,0,0,0,0,0,0};
int gaspointer = 0;

String livroom = "living_room";
String garden = "garden";
String garage = "garage";
String bedroom = "bedroom";
String toilet = "toilet";
String kitchen = "kitchen";
String base_url = "https://embeddedhardware-production.up.railway.app/room/";
String current_room = "living_room";
String current_url = base_url + current_room;

bool lightwarnstate = false;
bool airconstate = false;
bool isgaswarn = false;
bool isairon = false;
bool isstopsignon=false;

void ServoController(int servo_num, int servo_pos);
void ledController(int led_num, int status);
void motorController(bool status);
long ReadSonicSensor();
void turnaircon();
void updateaircon(String current_url);
void poolrq(String current_url);
void poolgasrq(String current_url);
void setuproomdoc();
void ConnectWifi();
void TVController(int status);
float ReadTemperature();
int ReadGasSensor();
long microsecondsToCentimeters(long microseconds);
int true21(bool status);

struct HomeEnvironment {
  bool light;
  bool door;
  bool airConditioner;
  bool temperature;
  bool gasDetector;
  bool fan;
};

StaticJsonDocument<512> roomdoc; //check room detail ex. living_room hava light, door,etc.
// StaticJsonDocument<512> roomdata; //check room's current status ex. light on, door open,etc.

TaskHandle_t Task1;
TaskHandle_t Task2;

void setup() {
  Serial.begin(115200);
  pinMode(temp_pin, INPUT);
  pinMode(motor_pin, OUTPUT);
  pinMode(gas_pin, INPUT);
  Serial.println("Inside setup");
  setuproomdoc();
  if(roomdoc[current_room]["temperature"]){
    pinMode(temp_pin, INPUT);
  }
  if(roomdoc[current_room]["gasDetector"]){
    pinMode(gas_pin, INPUT);
  }
  if(roomdoc[current_room]["fan"]){
    pinMode(motor_pin, OUTPUT);
  }
  if(roomdoc[current_room]["light"]){
    for (int i = 0;i < 2;i++) {
      pinMode(led_pin[i], OUTPUT);
    }
  }
  if(roomdoc[current_room]["door"]){
    pinMode(servo_pin[0], OUTPUT);
    servo01.attach(servo_pin[0]);
  }
  if(roomdoc[current_room]["airConditioner"]){
    pinMode(servo_pin[1], OUTPUT);
    servo02.attach(servo_pin[1]);
  }
  if(current_room == garage){
    pinMode(pingPin, OUTPUT);
    pinMode(inPin, INPUT);
  }

  ConnectWifi();
  analogReadResolution(12);

  if (!display.begin(SSD1306_SWITCHCAPVCC, 0x3C)) {
    Serial.println("SSD1306 allocation failed");
  }
  else {
    Serial.println("display Start Work !!!");
  }

  display.clearDisplay();

  xTaskCreatePinnedToCore(
    request,   /* Task function. */
    "request",     /* name of task. */
    10000,       /* Stack size of task */
    NULL,        /* parameter of the task */
    1,           /* priority of the task */
    &Task1,      /* Task handle to keep track of created task */
    0);          /* pin task to core 0 */

  //create a task that will be executed in the Task2code() function, with priority 1 and executed on core 1
  xTaskCreatePinnedToCore(
    nonrequest,   /* Task function. */
    "nonrequest",     /* name of task. */
    10000,       /* Stack size of task */
    NULL,        /* parameter of the task */
    1,           /* priority of the task */
    &Task2,      /* Task handle to keep track of created task */
    1);          /* pin task to core 1 */
}

void request(void* parameter) {
  for (;;) {
    poolrq(current_url);
    delay(1000);
  }
}

void nonrequest(void* parameter) {
  for (;;) {
    if (roomdoc[current_room]["airConditioner"] && isairon) {
      turnaircon();
    }
    if (isgaswarn) {
      if (!lightwarnstate) {
        ledController(2, 1);
        lightwarnstate = true;
      }
      else if (lightwarnstate) {
        ledController(2, 0);
        lightwarnstate = false;
      }
    }
    if (isstopsignon) {
      if (!lightwarnstate) {
        TVController(2);
        lightwarnstate = true;
      }
      else if (lightwarnstate) {
        TVController(0);
        lightwarnstate = false;
      }
    }
    delay(500);
  }
}

void loop() {
  // long sonic_value = ReadSonicSensor();
  // Serial.print(sonic_value);
  // Serial.println(" cm");
  // if(sonic_value<5){
  //   nearsonicounter++;
  // }else if(sonic_value>=5){
  //   nearsonicounter=0;
  //   TVController(0);
  // }
  // if(nearsonicounter>5){
  //   TVController(2);
  // }
  // for (int pos = 0; pos <= 180; pos += 1) { // goes from 0 degrees to 180 degrees
  //   ServoController(0, pos);
  //   delay(100);(15);                       // waits 15ms for the servo to reach the position
  // }
  // for (int pos = 180; pos >= 0; pos -= 1) { // goes from 180 degrees to 0 degrees
  //   ServoController(1, pos);
  //   delay(100);(15);                       // waits 15ms for the servo to reach the position
  // }
  // PT_SCHEDULE(slowBlinkThread(&ptSlowBlink));
  // if (roomdoc[current_room]["airConditioner"] && isairon) {
  //   turnaircon();
  // }
  // poolrq(current_url);
  // if(isgaswarn){
  //   if(!lightwarnstate){
  //     ledController(2, 1);
  //     lightwarnstate=true;
  //   }else if(lightwarnstate){
  //     ledController(2, 0);
  //     lightwarnstate=false;
  //   }
  // }
  // if (roomdoc[current_room]["airConditioner"] && isairon) {
  //   turnaircon();
  // }
}

void ConnectWifi() {
  WiFi.begin(ssid, password);

  while (WiFi.status() != WL_CONNECTED) {
    delay(100);
    Serial.print(".");
  }

  Serial.println("");
  Serial.println("WiFi connected");
  Serial.println("IP address: ");
  Serial.println(WiFi.localIP());
}

void turnaircon() {
  if (airconstate) {
    ServoController(1, 10);
    airconstate = false;
  }
  else if (!airconstate) {
    ServoController(1, 40);
    airconstate = true;
  }
}

// int slowBlinkThread(struct pt* pt) {
//   PT_BEGIN(pt);

//   while (isgaswarn) {
//     // Serial.println(".");
//     ledController(2, 1);
//     PT_SLEEP(pt, 500);
//     ledController(2, 0);
//     PT_SLEEP(pt, 500);
//   }

//   PT_END(pt);
// }

void ServoController(int servo_num, int servo_pos) {
  // servo_num is 0, 1 and 2 for all
  // servo_pos in [0, 180]
  switch (servo_num) {
  case 0:
    servo01.write(servo_pos);
    break;
  case 1:
    servo02.write(servo_pos);
    break;
  case 2:
    servo01.write(servo_pos);
    servo02.write(servo_pos);
    break;
  }
}

void ledController(int led_num, int status) {
  // led_num is 0, 1 and 2 for all
  // status is 0 or 1
  if (led_num == 2) {
    digitalWrite(led_pin[0], status);
    digitalWrite(led_pin[1], status);
  }
  else if (led_num == 0 || led_num == 1) {
    digitalWrite(led_pin[led_num], status);
  }
}

void motorController(bool status) {
  digitalWrite(motor_pin, status);
}

float ReadTemperature() {
  int adc = analogRead(temp_pin);
  float Beta = 3928.55;
  float Rat25C = 10000;
  float Rone = 10000;
  float RatInf = Rat25C * exp(-Beta / (25 + 273));
  float RatTemp = Rone * (4096 - adc) / adc;
  float Temperature_K = Beta / log(RatTemp / RatInf);
  float Temperature_C = Temperature_K - 273;
  return Temperature_C;
}

long ReadSonicSensor() {
  digitalWrite(pingPin, LOW);
  delayMicroseconds(2);
  digitalWrite(pingPin, HIGH);
  delayMicroseconds(10);
  digitalWrite(pingPin, LOW);
  long duration = pulseIn(inPin, HIGH);
  return microsecondsToCentimeters(duration);
}

int ReadGasSensor() {
  return analogRead(gas_pin);
}

void TVController(int status) {
  display.clearDisplay();
  if (status == 1) {
    display.fillScreen(SSD1306_WHITE);
  }else if (status == 0){
    display.fillScreen(SSD1306_BLACK);
  }else if (status == 2){
    display.setTextSize(3);
    display.setTextColor(WHITE);
    display.setCursor(0, 10);
    // Display static text
    display.println("Stop!!");
  }
  display.display();
}

long microsecondsToCentimeters(long microseconds)
{
// The speed of sound is 340 m/s or 29 microseconds per centimeter.
// The ping travels out and back, so to find the distance of the
// object we take half of the distance travelled.
return microseconds / 29 / 2;
}

int findgasmean(){
  int sum=0;
  for(int i=0;i<10;i++){
    sum+=gas_array[i];
  }
  return sum/10;
}

void poolrq(String current_url) {
  Serial.println(current_room);
  if (roomdoc[current_room]["gasDetector"]) {
    poolgasrq(base_url);
  }
  if (WiFi.status() == WL_CONNECTED) {
    HTTPClient http;

    // Specify the target URL
    http.begin(current_url);

    // Make a GET request
    int httpResponseCode = http.GET();

    if (httpResponseCode > 0) {
      Serial.print("HTTP Response Code: ");
      Serial.println(httpResponseCode);

      // Parse JSON
      DynamicJsonDocument jsonBuffer(1024);
      deserializeJson(jsonBuffer, http.getString());

      // Access parsed values
      bool lightstatus = jsonBuffer["service"]["light"];
      bool isdooropen = jsonBuffer["service"]["door"];
      isairon = jsonBuffer["service"]["airConditioner"];
      bool gasdetect = jsonBuffer["service"]["gasDetector"];
      bool fanstatus = jsonBuffer["service"]["fan"];
      Serial.println("Parsed JSON Value: ");
      // Serial.'println'

      isgaswarn = jsonBuffer["service"]["gasDetector"];
      Serial.print("Warning Signal: ");
      Serial.println(isgaswarn);
      // if(isgaswarn){

      //   if(!lightwarnstate){
      //     ledController(2, 1);
      //     lightwarnstate=true;
      //   }else if(lightwarnstate){
      //     ledController(2, 0);
      //     lightwarnstate=false;
      //   }
      // }  

      if (roomdoc[current_room]["light"]) {
        Serial.print("Light: ");
        Serial.println(lightstatus);
        if (!isgaswarn) {
          ledController(2, lightstatus);
        }
      }
      if (roomdoc[current_room]["door"]) {
        int door_value = 0;
        Serial.print("Door: ");
        Serial.println(isdooropen);
        if (isdooropen) {
          door_value = 90;
        }
        ServoController(0, door_value);
      }
      if (roomdoc[current_room]["airConditioner"]) {
        Serial.print("Aircon: ");
        Serial.println(isairon);
        if (isairon) {
          turnaircon();
        }
        if (roomdoc[current_room]["temperature"]) {
          Serial.print("Temperature: ");
          Serial.println(ReadTemperature());
          if (ReadTemperature() > 30 && !isairon) {
            updateaircon(base_url);
            turnaircon();
          }
        }
      }
      if (roomdoc[current_room]["fan"]) {
        Serial.print("Fan: ");
        Serial.println(fanstatus);
        motorController(fanstatus);
      }
      if (roomdoc[current_room]["Tv"]) {
        bool tvstatus = jsonBuffer["service"]["Tv"];
        Serial.print("Tv: ");
        Serial.println(tvstatus);
        TVController(tvstatus);
      }
      if (current_room == garage){
        long sonic_value = ReadSonicSensor();
        Serial.print("Sonic: ");
        Serial.print(sonic_value);
        Serial.println(" cm");
        if(sonic_value<10){
          nearsonicounter++;
        }else if(sonic_value>=10){
          nearsonicounter=0;
          isstopsignon=false;
          TVController(0);
        }
        if(nearsonicounter>2){
          isstopsignon=true;
          TVController(2);
        }

      }
    }
    else {
      Serial.print("HTTP GET request failed. Error code: ");
      Serial.println(httpResponseCode);
    }

    http.end();

    // delay(100);(500); // Delay for 5 seconds before making another request
  }
}

void poolgasrq(String base_url) {
  int gas_value = ReadGasSensor();
  gas_array[gaspointer] = gas_value;
  gaspointer++;

  if(gaspointer>9){
    gaspointer=0;
  }

  int gasmean = findgasmean();

  if (gasmean > 750) {
    gascounter++;
  }
  else if (gasmean <= 750) {
    gascounter = 0;
  }

  if (gascounter > 5 && !isgaswarn) {
    if (WiFi.status() == WL_CONNECTED) {
      HTTPClient http;
      http.begin(base_url + "edit");
      DynamicJsonDocument edit_body(1024);
      edit_body["service"]["gasDetector"] = true;
      edit_body["name"] = current_room;
      String edit_json;
      serializeJson(edit_body, edit_json);
      http.addHeader("Content-Type", "application/json");
      int httpResponseCode = http.PUT(edit_json);
      if (httpResponseCode > 0) {
        isgaswarn = true;
        Serial.print("HTTP Response Code: ");
        Serial.println(httpResponseCode);
      }
      else {
        Serial.print("HTTP PUT request failed. Error code: ");
        Serial.println(httpResponseCode);
        Serial.println("Try again....");
        poolgasrq(base_url);
      }
      delay(100); // Delay for 0.1 seconds before making another request
    }
  }
  else if (gasmean <= 750 && isgaswarn) {
    gascounter = 0;
    if (WiFi.status() == WL_CONNECTED) {
      HTTPClient http;
      http.begin(base_url + "edit");
      DynamicJsonDocument edit_body(1024);
      edit_body["service"]["gasDetector"] = false;
      edit_body["name"] = current_room;
      String edit_json;
      serializeJson(edit_body, edit_json);
      http.addHeader("Content-Type", "application/json");
      int httpResponseCode = http.PUT(edit_json);
      if (httpResponseCode > 0) {
        isgaswarn = false;
        Serial.print("HTTP Response Code: ");
        Serial.println(httpResponseCode);
      }
      else {
        Serial.print("HTTP PUT request failed. Error code: ");
        Serial.println(httpResponseCode);
        Serial.println("Try again....");
        poolgasrq(base_url);
      }
      delay(100); // Delay for 0.1 seconds before making another request
    }
  }
}

void updateaircon(String base_url) {
  if (WiFi.status() == WL_CONNECTED) {
    HTTPClient http;
    http.begin(base_url + "edit");
    DynamicJsonDocument edit_body(1024);
    edit_body["service"]["airConditioner"] = true;
    edit_body["name"] = current_room;
    String edit_json;
    serializeJson(edit_body, edit_json);
    http.addHeader("Content-Type", "application/json");
    int httpResponseCode = http.PUT(edit_json);
    if (httpResponseCode > 0) {
      Serial.print("HTTP Response Code: ");
      Serial.println(httpResponseCode);
    }
    else {
      Serial.print("HTTP PUT request failed. Error code: ");
      Serial.println(httpResponseCode);
      Serial.println("Try again....");
    }
    delay(100); // Delay for 0.1 seconds before making another request
  }
}

int true21(bool status) {
  if (status) {
    return 1;
  }
  return 0;
}

void setuproomdoc() {
  roomdoc["living_room"]["light"] = true;
  roomdoc["living_room"]["door"] = true;
  roomdoc["living_room"]["airConditioner"] = false;
  roomdoc["living_room"]["temperature"] = false;
  roomdoc["living_room"]["gasDetector"] = false;
  roomdoc["living_room"]["fan"] = true;
  roomdoc["living_room"]["Tv"] = true;

  roomdoc["garden"]["light"] = true;
  roomdoc["garden"]["door"] = true;
  roomdoc["garden"]["airConditioner"] = false;
  roomdoc["garden"]["temperature"] = false;
  roomdoc["garden"]["gasDetector"] = false;
  roomdoc["garden"]["fan"] = false;
  roomdoc["garden"]["Tv"] = false;

  roomdoc["garage"]["light"] = true;
  roomdoc["garage"]["door"] = false;
  roomdoc["garage"]["airConditioner"] = false;
  roomdoc["garage"]["temperature"] = false;
  roomdoc["garage"]["gasDetector"] = false;
  roomdoc["garage"]["fan"] = false;
  roomdoc["garage"]["Tv"] = false;

  roomdoc["bedroom"]["light"] = true;
  roomdoc["bedroom"]["door"] = true;
  roomdoc["bedroom"]["airConditioner"] = true;
  roomdoc["bedroom"]["temperature"] = true;
  roomdoc["bedroom"]["gasDetector"] = false;
  roomdoc["bedroom"]["fan"] = false;
  roomdoc["bedroom"]["Tv"] = false;

  roomdoc["toilet"]["light"] = true;
  roomdoc["toilet"]["door"] = true;
  roomdoc["toilet"]["airConditioner"] = false;
  roomdoc["toilet"]["temperature"] = false;
  roomdoc["toilet"]["gasDetector"] = false;
  roomdoc["toilet"]["fan"] = false;
  roomdoc["toilet"]["Tv"] = false;

  roomdoc["kitchen"]["light"] = true;
  roomdoc["kitchen"]["door"] = true;
  roomdoc["kitchen"]["airConditioner"] = false;
  roomdoc["kitchen"]["temperature"] = false;
  roomdoc["kitchen"]["gasDetector"] = true;
  roomdoc["kitchen"]["fan"] = false;
  roomdoc["kitchen"]["Tv"] = false;
}

// {"living_room":{
//     "light":true,
//     "door":true,
//     "airConditioner":false,
//     "temperature":false,
//     "gasDetector":false,
//     "fan":true
// },"garden":{
//     "light":true,
//     "door":true,
//     "airConditioner":false,
//     "temperature":false,
//     "gasDetector":false,
//     "fan":false
// },"garage":{
//     "light":true,
//     "door":true,
//     "airConditioner":false,
//     "temperature":false,
//     "gasDetector":false,
//     "fan":false
// },"bedroom":{
//     "light":true,
//     "door":true,
//     "airConditioner":true,
//     "temperature":false,
//     "gasDetector":false
// },"toilet":{
//     "light":true,
//     "door"true,
//     "airConditioner":false,
//     "temperature":false,
//     "gasDetector":false
// },"kitchen":{
//     "light":true,
//     "door"true,
//     "airConditioner":false,
//     "temperature":false,
//     "gasDetector":true
// }}
