import {
  Document,
  Model,
  model,
  Schema,
  SchemaTimestampsConfig,
} from 'mongoose'
import { uuidv4 } from '@firebase/util'

export interface Service {
  door?: boolean
  light?: boolean
  airConditioner?: boolean
  temperature?: number
  gasDetector?: boolean
  [key: string]: any
}
export interface IRoom {
  _id: string
  name: string
  service: Service
}

interface IService extends Service,Document, SchemaTimestampsConfig {}
const  serviceSchema = new Schema<IService,IService>({
  door:{
    type:Boolean,
    default:false
  },
  light:{
    type:Boolean,
    default:false
  },
  airConditioner:{
    type:Boolean,
    default:false
  },
  temperature:{
    type:Number,
    default:0
  },
  gasDetector:{
    type:Boolean,
    default:false
  }
},{_id:false,strict:false})

interface IRoomDocument extends IRoom, Document, SchemaTimestampsConfig {
  _id: string
}

type IRoomModel = Model<IRoomDocument>

const roomSchema = new Schema<IRoomDocument, IRoomDocument>(
  {
    _id: {
      type: String,
      default: () => uuidv4(),
    },
    name: {
      type: String,
      required: true,
    },
    service:{
      type:serviceSchema,
      default:{}
    }
  },
  { timestamps: true },
)

export default model<IRoomDocument, IRoomModel>('Room', roomSchema)
