import { Router } from "express";
import { ICommandVector } from "../models/commandVector";
const router = Router();
import openai from "../share/gpt";
import { Embeddings } from "openai/resources";
import CommandVector from "../models/commandVector";
import { Service } from "../models/room";
interface ICommandCreateVector {
    command: string,
    room: string,
    controll: Service
}

router.post('/create', async (req, res) => {
    try {
     const { command,room,controll } = <ICommandCreateVector>req.body
    //  console.log(command,room,controll)
     const embedding = await openai.embeddings.create({
        model: "text-embedding-3-small",
        input: command,
        encoding_format: "float",
      });
    const vector = embedding.data[0].embedding
    const commandVector = await CommandVector.create(
        {command, vector, room, controll}
    )
    if (commandVector) {
        return res.status(200).send(commandVector)
    }
      return res.status(500).send('Failed to create embedding')
    } catch (error) {
      return res.status(500).send(error)
    }
  })
router.post('/search', async (req, res) => {
    try {
     const { command } = <ICommandCreateVector>req.body
    //  console.log(command)
     const embedding = await openai.embeddings.create({
        model: "text-embedding-3-small",
        input: command,
        encoding_format: "float",
      });
     
    // console.log(embedding.data[0].embedding)
    const pipe = [
      {
        "$vectorSearch": {
          index: 'vector_command',
          path: 'vector',
          queryVector: embedding.data[0].embedding, // Adjust the query vector as needed
          numCandidates:  30,
          limit:1 ,
        },
      },
      {
        "$project": {
          _id: 1,
          command: 1,
          room: 1,
          controll: 1,
          'score': {
            '$meta': 'vectorSearchScore'
          },
          
          
        },
        
      },
      {
        "$match": {
          "score": { 
            "$gte": 0.9
         }
        }
      }
    ] as any

    const commandMatch = await CommandVector.aggregate(pipe)
    // console.log(commandMatch)
    if (commandMatch) {
        return res.status(200).send(commandMatch)
    }
      return res.status(500).send('Failed to create embedding')
    } catch (error) {
      console.log(error)
      return res.status(500).send(error)
    }
  })

export default router
  